import pytest
import torch

from rcmodel.tools import InputScaling


@pytest.fixture
def scaling():
    rm_CA = [200, 800]  # [min, max] Capacitance/area
    ex_C = [1.5 * 10 ** 4, 10 ** 6]  # Capacitance
    R = [0.2, 1.2]  # Resistance ((K.m^2)/W)
    Q_limit = [-300, 500]  # gain limit (W/m^2)

    scaling = InputScaling(rm_CA, ex_C, R, Q_limit)
    return scaling


def test_physical_param_scaling(scaling):
    theta = 0.5 * torch.ones(scaling.get_n_params())
    theta_physical = scaling.physical_param_scaling(theta)
    print(theta_physical)
    assert torch.equal(theta_physical, torch.tensor(
        [5.0000e+02, 5.0750e+05, 5.0750e+05, 7.0000e-01, 7.0000e-01, 7.0000e-01, 7.0000e-01, 1.0000e+02]))


def test_model_param_scaling(scaling):
    model_scaled = scaling.model_param_scaling(
        torch.tensor([5.0000e+02, 5.0750e+05, 5.0750e+05, 7.0000e-01, 7.0000e-01, 7.0000e-01, 7.0000e-01, 1.0000e+02]))

    assert torch.equal(model_scaled, 0.5 * torch.ones(scaling.get_n_params()))


def test_physical_cooling_scaling(scaling):
    cool_physical = scaling.physical_cooling_scaling(0.2)

    assert cool_physical == 100


def test_model_cooling_scaling(scaling):
    cool_scaled = scaling.model_cooling_scaling(100)

    assert cool_scaled == 0.2


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# coding: utf-8

# In[1]:


import pandas as pd
import matplotlib.pyplot as plt
import torch
import numpy as np
from pathlib import Path

from torch import Tensor
from tqdm.auto import tqdm, trange

from rcmodel import *
from main import initialise_model, RayActor


# In[2]:


torch.cuda.is_available = lambda: False


# In[3]:


import gym
from gym import spaces

class LSIEnv(gym.Env):
    """Custom Environment that follows gym interface"""
    metadata = {'render.modes': ['human']}

    def __init__(self, RC, time_data):
        super().__init__()

        self.RC = RC  # RCModel Class
        self.time_data = time_data  # timeseries
        self.t_index = 0  # used to keep track of index through timeseries

        # ----- GYM Stuff -----
        self.n_rooms = len(self.RC.building.rooms)
        self.low_state = -10
        self.high_state = 50
        # Define action and observation space
        # They must be gym.spaces objects
        # Example when using discrete actions:
        self.action_space = spaces.Discrete(2, )

        # Observation is temperature of each room.
        self.observation_space = spaces.Box(
            low=self.low_state,
            high=self.high_state,
            shape=(self.n_rooms,),
            dtype=np.float32
        )

#                  action
    def step(self, step_size):
        # Execute a chunk of timeseries
        t_eval = self.time_data[self.t_index:int(self.t_index + step_size)]

        # actions are decided and stored by the policy while integrating the ODE:
        pred = self.RC(t_eval)
        
        return pred.detach()  # No need for grad
    #          (observation, reward, done, info)
    # self.state, reward, done, {}

    def reset(self):
        # Reset the state of the environment to an initial state
        self.t_index = 0
        self.RC.reset_iv()

        self.RC.cooling_policy.on_policy_reset()

    def render(self, mode='human', close=False):
        # Render the environment to the screen

        return


# In[4]:




opt_id = 0

weather_data_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/Met Office Weather Files/JuneSept.csv'
csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/test2d_sorted.csv'
#
# weather_data_path = '/home/benf/LSI/Data/Met Office Weather Files/JuneSept.csv'
# csv_path = '/home/benf/LSI/Data/DummyData/train5d_sorted.csv'  # where building data
dir_path = f'./outputs/run{opt_id}/models/'  # where to save
# Load model?
load_model_path = None  # or None
start_num = 0  # Number of cycles to start at. Used if resuming the run. i.e. the first cycle is (start_num + 1)



def do_plots(ER_plot, rewards_plot, opt_id):
    
    fig, axs = plt.subplots(1, 2, figsize=(10, 7), dpi=400)
    y = torch.flatten(torch.tensor(ER_plot)).detach().numpy()
    axs[0].plot(range(1, len(y) + 1), y, 'b', label='expected rewards')
    axs[0].legend()

    y = torch.flatten(torch.tensor(rewards_plot)).detach().numpy()
    axs[1].plot(range(1, len(y) + 1), y, 'r', label='total rewards')
    axs[1].legend()
    fig.suptitle('Rewards', fontsize=16)
    axs[0].set_xlabel('Epoch')
    axs[1].set_xlabel('Epoch')
    axs[0].set_ylabel('Reward')
    axs[1].set_ylabel('Reward')

    fig.savefig(f'./outputs/run{opt_id}/plots/RewardsPlot.png')
    plt.close()


def init_scaling():
    # Initialise scaling class
    rm_CA = [100, 1e4]  # [min, max] Capacitance/area
    ex_C = [1e3, 1e8]  # Capacitance
    R = [0.1, 5]  # Resistance ((K.m^2)/W)
    scaling = InputScaling(rm_CA, ex_C, R)
    return scaling


# Initialise Model
scaling = init_scaling()

# Initialise Optimise Class - for training physical model
dt = 30
sample_size = 24 * 60 ** 2 / dt

# Initialise Reinforcement Class - for training policy
time_data = torch.tensor(pd.read_csv(csv_path, skiprows=0).iloc[:, 1], dtype=torch.float64)
temp_data = torch.tensor(pd.read_csv(csv_path, skiprows=0).iloc[:, 2:].to_numpy(dtype=np.float32), dtype=torch.float32)


# Dataloader to be used when plotting a model run. This is just for info.
plot_results_data = tools.BuildingTemperatureDataset(csv_path, 5 * sample_size, all=True)
plot_dataloader = torch.utils.data.DataLoader(plot_results_data, batch_size=1, shuffle=False)






# In[ ]:





# In[16]:


def train_policy(rl, opt_id):
    # lists to keep track of process, used for plot at the end
    avg_train_loss_plot = []
    avg_test_loss_plot = []
    rewards_plot = []
    ER_plot = []
    count = 1  # Counts number of epochs since start.


    # check if dir exists and make if needed
    Path(f'./outputs/run{opt_id}/plots/results/').mkdir(parents=True, exist_ok=True)

    # Convergence happens when the mean gradient of loss/reward is < tol.
    tol_policy = 50


    cycles = 1
    max_epochs = 200

    plt.ioff()  # Reduces memory usage by matplotlib
    for cycle in range(cycles):
        # -------- Policy training --------
        tqdm.write('Policy Training:')
        diff = torch.ones(5) * 1000
        rewards_prev = 0
        for epoch in trange(max_epochs):
            rewards, ER = rl.train(1, sample_size)

            indx = epoch % len(diff)
            diff[indx] = abs(rewards_prev - torch.tensor(rewards))  # keeps track of a window of differences
            rewards_prev = torch.tensor(rewards)
            rewards_plot.append(rewards)
            ER_plot.append(ER)

            tqdm.write(
                f'Epoch {count}, Rewards/Expected Rewards: {torch.tensor(rewards).sum().item():.2f}/{torch.tensor(ER).sum().item():.2f}, Mean diff: {torch.mean(diff):.2f}')

            # Save Model
            model_id = count + start_num
            rl.env.RC.save(model_id, dir_path)
            count += 1

#             if torch.mean(diff) < tol_policy:
#                 break

        tqdm.write(f'Policy converged in {epoch + 1} epochs. Total epochs: {count - 1}\n')

        # Save a plot of results after policy training
        pltsolution_1rm(rl.env.RC, plot_dataloader, f'./outputs/run{opt_id}/plots/results/Result_Cycle{start_num + cycle + 1}b.png')

        # Plot loss and reward plot
        do_plots(ER_plot, rewards_plot, opt_id)


# In[ ]:





# In[6]:


class Reinforce_baseline:
    def __init__(self, env, time_data, temp_data, gamma=0.9999, alpha=1e-3):

        assert len(time_data) == len(temp_data)

        self.env = env
        # self.pi = pi
        self.time_data = time_data
        self.temp_data = temp_data
        self.gamma = gamma
        self.alpha = alpha
        self.baseline = 0

        self.optimiser = torch.optim.Adam(self.env.RC.cooling_policy.parameters(), lr=self.alpha)

    def update_policy(self, rewards, log_probs):

        # Calculate Discounted Reward:
        discounted_rewards = torch.zeros(len(rewards))

        R = 0
        indx = len(rewards) - 1
        for r in reversed(rewards):
            R = r + self.gamma * R  # Discounted Reward is calculated from last reward to first.

            discounted_rewards[indx] = R  # Fill array back to front to un-reverse the order
            indx -= 1

        expected_reward = -torch.stack(log_probs) * (discounted_rewards - self.baseline)  # negative for maximising
        expected_reward = torch.sum(expected_reward)

        # Update parameters in pi
        self.optimiser.zero_grad()
        expected_reward.backward()
        self.optimiser.step()

        # print(list(self.pi.parameters())[0].grad)  # check on grads if needed

        return expected_reward, discounted_rewards

    def train(self, num_episodes, step_size):
        self.env.RC.cooling_policy.train()  # Put in training mode
        total_ER = []
        total_rewards = []

        loss_fn = torch.nn.MSELoss(reduction='none')  # Squared Error

        # with tqdm(total=len(self.env.time_data) * num_episodes, position=0, leave=False) as pbar:  # progress bar
        for episode in range(num_episodes):
            self.env.reset()

            episode_rewards = []
            episode_ER = []

            # Time is increased in steps, with the policy updating after every step.
            n_steps = int( np.ceil(len(self.env.time_data) / step_size) )
            
            for step_count in range(1, n_steps+1):
                
                # takes a step_size forward in time
                pred = self.env.step(step_size).squeeze(-1)  # state and action produced in step
                
                actual = self.temp_data[self.env.t_index:int(self.env.t_index + step_size), 0:self.env.n_rooms]

                # negative so reward can be maximised
                reward = -loss_fn(pred[:, 2:], actual)

                # Do gradient decent on sample
                ER, DR = self.update_policy(reward, self.env.RC.cooling_policy.log_probs)

                self.env.RC.cooling_policy.on_policy_reset()  # empty buffer

                # get last output and use for next initial value
                self.env.RC.iv = pred[-1, :].unsqueeze(1).detach()  # MUST DETACH GRAD

                episode_rewards.append(sum(reward))
                episode_ER.append(ER)

                self.env.t_index += int(step_size)  # increase environment time

                # update baseline as the mean of all discounted rewards in episode
                if self.baseline == 0:  # initial case is different
                    self.baseline = DR.mean() / step_count
                else:
                    self.baseline = (self.baseline*step_count + DR.mean()) / (step_count+1)
            
            # At end of episode set baseline to mean of final discounted rewards.
            # Carries over some info between episodes without
            self.baseline = DR.mean()

#             print(f'Episode {episode+1}, Expected Reward: {sum(episode_ER).item():.2f}, total_reward: {sum(episode_rewards).item():.2f}')

            total_ER.append(sum(episode_ER).detach())
            total_rewards.append(sum(episode_rewards).detach())

        return total_rewards, total_ER


# In[7]:


for i in range(1):
    policy = PolicyNetwork(7, 2)
    model = initialise_model(policy, scaling, weather_data_path)
    if load_model_path:
        model.load(load_model_path)# Reset loaded cooling policy with a blank one
        model.cooling_policy = PolicyNetwork(7, 2)
    env = LSIEnv(model, time_data)
    rl_b = Reinforce_baseline(env, time_data, temp_data)
    train_policy(rl_b, f'{opt_id}{i}')


# In[8]:


# for i in range(5):
#     policy = PolicyNetwork(7, 2)
#     model = initialise_model(policy, scaling, weather_data_path)
#     if load_model_path:
#         model.load(load_model_path)# Reset loaded cooling policy with a blank one
#         model.cooling_policy = PolicyNetwork(7, 2)
#     env = LSIEnv(model, time_data)
#     opt_id = 'NoBaseline'


#     rl = Reinforce(env, time_data, temp_data, alpha=1e-2)
#     train_policy(rl, f'{opt_id}{i}')


# In[9]:


import ray
import copy
import time

def get_rl_baseline():
    policy = PolicyNetwork(7, 2)
    model = initialise_model(policy, scaling, weather_data_path)
    if load_model_path:
        model.load(load_model_path)# Reset loaded cooling policy with a blank one
        model.cooling_policy = PolicyNetwork(7, 2)
#     env = LSIEnv(copy.deepcopy(model), time_data)
    
#     return Reinforce_baseline(copy.deepcopy(env), time_data, temp_data)
    env = LSIEnv(model, time_data)
    
    return Reinforce_baseline(env, time_data, temp_data)

def get_rl():
    policy = PolicyNetwork(7, 2)
    model = initialise_model(policy, scaling, weather_data_path)
    if load_model_path:
        model.load(load_model_path)# Reset loaded cooling policy with a blank one
        model.cooling_policy = PolicyNetwork(7, 2)
    env = LSIEnv(model, time_data)
    opt_id = 'NoBaseline'
    return Reinforce(env, time_data, temp_data, alpha=1e-2)



train_policy = ray.remote(train_policy)  # Make remote function


trials = []
opt_id = []
for i in range(5):
    trials.append(get_rl_baseline())
    opt_id.append(f'Baseline{i}')

for i in range(5):
    trials.append(get_rl())
    opt_id.append(f'NoBaseline{i}')


# train_policy(trials[0], f'{opt_id[0]}')
actors = [train_policy.remote(rl, f'{opt_id[i]}') for i, rl in enumerate(trials)]


results = ray.get([a for num, a in enumerate(actors)])

time.sleep(3)  # Print is cut short without sleep

ray.shutdown()


# In[13]:


trials[0].env.RC.cooling_policy.state_dict()['linear_relu_stack.0.weight']


# In[14]:


trials[1].env.RC.cooling_policy.state_dict()['linear_relu_stack.0.weight']


# In[10]:


ray.shutdown()


# In[ ]:





# In[ ]:







from filelock import FileLock
import torch
import pandas as pd
from .tools import BuildingTemperatureDataset
import os


def train(model, dataloader, optimizer):
    """
    Performs one epoch of training.
    Order of rooms in building and in data must match otherwise model will fit wrong rooms to data.
    """
    model.reset_iv()  # Reset initial value
    model.train()
    model.cooling_policy.eval()

    # Stops Autograd endlessly keeping track of the graph. Memory Leak!
    for layer in model.cooling_policy.parameters():
        layer.requires_grad = False

    num_cols = len(model.building.rooms)  # number of columns to use from data.
    num_batches = len(dataloader)
    train_loss = 0
    loss_fn = torch.nn.MSELoss()

    for batch, (time, temp) in enumerate(dataloader):
        # Get model arguments:
        time = time.squeeze(0)
        temp = temp.squeeze(0)

        # Compute prediction and loss
        pred = model(time)
        pred = pred.squeeze(-1)  # change from column to row matrix

        loss = loss_fn(pred[:, 2:], temp[:, 0:num_cols])
        train_loss += loss.item()

        # get last output and use for next initial value
        model.iv = pred[-1, :].unsqueeze(1).detach()  # MUST DETACH GRAD

        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    return train_loss / num_batches


def test(model, dataloader):
    model.reset_iv()  # Reset initial value
    model.eval()  # Put model in evaluation mode
    num_batches = len(dataloader)
    num_cols = len(model.building.rooms)  # number of columns to take from data.
    test_loss = 0
    loss_fn = torch.nn.MSELoss()

    with torch.no_grad():
        for (time, temp) in dataloader:

            time = time.squeeze(0)
            temp = temp.squeeze(0)

            pred = model(time)
            pred = pred.squeeze(-1)  # change from column to row matrix
            test_loss += loss_fn(pred[:, 2:], temp[:, 0:num_cols]).item()

            # get last output and use for next initial value
            model.iv = pred[-1, :].unsqueeze(1).detach()  # MUST DETACH GRAD

    test_loss /= num_batches

    return test_loss


def dataset_creator(path, sample_size, dt):
    path_sorted = sort_data(path, dt)
    with FileLock(f"{os.path.dirname(os.path.abspath(path_sorted))}.lock"):
        training_data = BuildingTemperatureDataset(path_sorted, sample_size, train=True)
        train_dataloader = torch.utils.data.DataLoader(training_data, batch_size=1, shuffle=False)
        test_data = BuildingTemperatureDataset(path_sorted, sample_size, test=True)
        test_dataloader = torch.utils.data.DataLoader(test_data, batch_size=1, shuffle=False)
        
    return train_dataloader, test_dataloader


def sort_data(path, dt):
    """
    Check if path has sorted data tag (_sorted)
    If not check if data has previously been sorted and exists in the directory.
    Check to see if the value dt is correct
    If not sort data and write filename_sorted.csv

    data is sorted by time in ascending order and downsampled to a frequency of dt seconds.
    Missing values are interpolated.
    A time-date string is also inserted.
    """
    def sort(path, dt):
        df = pd.read_csv(path)

        if path[-11:] == '_sorted.csv':
            path_sorted = path
        else:
            path_sorted = path[:-4] + '_sorted.csv'

        # Sort df by time (raw data not always in order)
        df = df.sort_values(by=["time"], ascending=True)

        # insert date-time value at start of df
        try:
            df.insert(loc=0, column='date-time', value=pd.to_datetime(df['time'], unit='ms'))
        except ValueError:
            raise ValueError('Data appears to have already been sorted. Check if still appropriate and add _sorted.csv tag to avoid this error.')

        # downscale data to a frequency of dt (seconds) use the mean value and round to 2dp.
        df = df.set_index('date-time').resample(str(dt) + 's').mean().round(2)

        # time column is converted to unix epoch seconds to match the date-time
        df["time"] = (df.index - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")

        # change date-time from UTC to Local time
        df = df.tz_localize('Europe/London')

        df = df.interpolate().round(2)  # interpolate missing values NaN

        df.to_csv(path_sorted, index=True)

    def need_to_sort(path, dt):

        def get_dt(path):
            df_dt = pd.read_csv(path)['time'][0:2].values
            return df_dt[1] - df_dt[0]

        # Does path already have sorted tag?
        if path[-11:] == '_sorted.csv':

            # if so, is dt correct?
            if get_dt(path) == dt:

                return False  # path and file is correct dont sort

            else:
                return True  # dt is wrong, re-sort

        # path does not contain _sorted.csv
        else:

            # Does path_sorted exist?
            path_sorted = path[:-4] + '_sorted.csv'
            import os.path
            if os.path.isfile(path_sorted):  # check if file already exists

                # if file exists check if dt is correct
                if get_dt(path_sorted) == dt:
                    return False  # correct file already exists don't sort
                else:
                    return True  # file exists but dt wrong, re-sort

            else: # File doesn't exist
                return True

    if need_to_sort(path, dt):
        sort(path, dt)

    # return the path_sorted
    if path[-11:] == '_sorted.csv':
        path_sorted = path
    else:
        path_sorted = path[:-4] + '_sorted.csv'

    return path_sorted


class OptimiseRC:
    """
    Parameters
    ----------
    model : object
        RCModel class object.
    csv_path : string
        Path to .csv file containing room temperature data.
        Data will be sorted if not done already and saved to a new file with the tag '_sorted'
    sample_size : int
        Length of indexes to sample from dataset per batch.
    dt : int
        Timestep data will be resampled to.
    lr : float
        Learning rate for optimiser.
    model_id : int
        Unique identifier used when optimising multiple models.

    see https://docs.ray.io/en/latest/using-ray-with-pytorch.html
    """
    def __init__(self, model, csv_path, sample_size, dt=30, lr=1e-3, opt_id=0):
        self.model = model
        self.model.init_params()  # randomise parameters
        self.model_id = opt_id
        self.train_dataloader, self.test_dataloader = dataset_creator(csv_path, int(sample_size), int(dt))

        self.optimizer = torch.optim.Adam([self.model.params, self.model.cooling], lr=lr)



    def train(self):
        avg_loss = train(self.model, self.train_dataloader, self.optimizer)
        return avg_loss

    def test(self):
        test_loss = test(self.model, self.test_dataloader)
        return test_loss

    def train_loop(self, epochs):
        print(self.model.params)
        for i in range(int(epochs)):
            # print(f"Epoch {i + 1}\n-------------------------------")
            testloss = self.train()

        results = [testloss, self.model]
        return results














import pytest

from rcmodel.building import Building
from rcmodel.room import Room


@pytest.fixture
def rooms_n2():
    """
    Two rooms side by side
    """
    rm1 = Room('rm1', [[0,0], [5,0], [5,5], [0,5]])
    rm2 = Room('rm2', [[5,0], [10,0], [10,5], [5,5]])
    rooms = [rm1, rm2]

    return rooms


@pytest.fixture
def rooms_n9():
    """
    Complicated arrangement of rooms
    """
    rooms = []

    rooms.append(Room('rm1', [[0,0], [5,0], [5,5], [0,5]]))
    rooms.append(Room('rm2', [[5,0], [10,0], [10,2], [10,4], [10,5], [5,5]]))
    rooms.append(Room('rm3', [[0,5], [5,5], [10,5], [10,6], [10,8], [10,10]]))

    rooms.append(Room('rm4', [[10,10], [12,10], [12,8], [10,8]]))
    rooms.append(Room('rm5', [[10,8], [12,8], [12,6], [10,6]]))
    rooms.append(Room('rm6', [[10,6], [12,6], [12,4], [10,4], [10,5]]))
    rooms.append(Room('rm7', [[10,4], [12,4], [12,2], [10,2]]))
    rooms.append(Room('rm8', [[10,2], [12,2], [12,0], [10,0]]))
    rooms.append(Room('rm9', [[12,0], [12,2], [12,4], [12,6], [12,8], [12,10], [14,10], [14,0]]))

    return rooms


@pytest.fixture
def building_n2(rooms_n2):
    height = 1
    rm_cap = [300]
    Ce = [1e3, 8e2]
    Re = [5, 1, 0.5]
    Rint = [0.1]
    gain = [0]

    theta = [rm_cap, Ce, Re, Rint, gain]
    theta = [item for sublist in theta for item in sublist]  # flatten list

    bld = Building(rooms_n2, height)
    bld.update_inputs(theta)

    return bld


@pytest.fixture
def building_n9(rooms_n9):
    height = 1
    rm_cap = [300]
    Ce = [1e3, 8e2]
    Re = [5, 1, 0.5]
    Rint = [0.1]
    gain = [0]

    theta = [rm_cap, Ce, Re, Rint, gain]
    theta = [item for sublist in theta for item in sublist]  # flatten list

    bld = Building(rooms_n9, height)
    bld.update_inputs(theta)

    return bld




import time
import torch
import pandas as pd
import numpy as np
import ray
from matplotlib import pyplot as plt
from tqdm.auto import tqdm, trange
from xitorch.interpolate import Interp1D
from pathlib import Path

from rcmodel import *
from rcmodel.optimisation import dataset_creator

from rcmodel.reinforce import PolicyNetwork
from rcmodel.reinforce import Reinforce
from rcmodel.reinforce import LSIEnv


def initialise_model(pi, scaling, weather_data_path):

    def change_origin(coords):
        x0 = 92.07
        y0 = 125.94

        for i in range(len(coords)):
            coords[i][0] = round((coords[i][0] - x0) / 10, 2)
            coords[i][1] = round((coords[i][1] - y0) / 10, 2)

        return coords

    rooms = []

    name = "seminar_rm_a_t0106"
    coords = change_origin(
        [[92.07, 125.94], [92.07, 231.74], [129.00, 231.74], [154.45, 231.74], [172.64, 231.74], [172.64, 125.94]])
    rooms.append(Room(name, coords))

    # Initialise Building
    bld = Building(rooms)

    df = pd.read_csv(weather_data_path)
    Tout = torch.tensor(df['Hourly Temperature (°C)'])
    t = torch.tensor(df['time'])

    Tout_continuous = Interp1D(t, Tout, method='linear')

    # Initialise RCModel with the building
    transform = torch.sigmoid
    model = RCModel(bld, scaling, Tout_continuous, transform, pi)

    return model


class RayActor:
    def __init__(self, scaling, weather_data_path, csv_path, ):
        self.scaling = scaling
        self.weather_data_path = weather_data_path
        self.csv_path = csv_path

    def worker(self, opt_id, epochs=100):
        torch.set_num_threads(1)
        # csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/train5d_sorted.csv'
        # csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/test2d_sorted.csv'
        # csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/210813data_sorted.csv'

        dir_path = f'./outputs/run{opt_id}/models/'  # where to save

        policy = PolicyNetwork(7, 2)
        model = initialise_model(policy, self.scaling, self.weather_data_path)

        model.save(0, dir_path)  # save initial model

        dt = 30
        sample_size = 24*60**2/dt

        op = OptimiseRC(model, self.csv_path, sample_size, dt, lr=1e-3, opt_id=opt_id)

        time_data = torch.tensor(pd.read_csv(self.csv_path, skiprows=0).iloc[:, 1], dtype=torch.float64)
        temp_data = torch.tensor(pd.read_csv(self.csv_path, skiprows=0).iloc[:, 2:].to_numpy(dtype=np.float32), dtype=torch.float32)

        env = LSIEnv(model, time_data)
        do_reinforce = Reinforce(env, time_data, temp_data, alpha=1e-2)

        avg_train_loss_plot = []
        avg_test_loss_plot = []
        rewards_plot = []
        ER_plot = []

        plot_results_data = tools.BuildingTemperatureDataset(self.csv_path, 5*sample_size, all=True)
        plot_dataloader = torch.utils.data.DataLoader(plot_results_data, batch_size=1, shuffle=False)

        # check if dir exists and make if needed
        Path(f'./outputs/run{opt_id}/plots/results/').mkdir(parents=True, exist_ok=True)

        for epoch in trange(epochs):
            # Physical Model Training:
            avg_train_loss = op.train()
            avg_train_loss_plot.append(avg_train_loss)

            # Policy Training:
            num_episodes = 1
            step_size = sample_size
            rewards, ER = do_reinforce.train(num_episodes, step_size)
            rewards_plot.append(rewards)
            ER_plot.append(ER)

            # Test Loss
            avg_test_loss = op.test()
            avg_test_loss_plot.append(avg_test_loss)

            # Save a plot of results every 10 epochs or if first/final epoch
            if (epoch+1) % 10 == 0 or epoch == epochs-1 or epoch == 0:
                pltsolution_1rm(model, plot_dataloader, f'./outputs/run{opt_id}/plots/results/Result_Epoch{epoch+1}.png')

            # Save Model
            model_id = epoch+1
            model.save(model_id, dir_path)

            tqdm.write(f'Epoch {epoch+1}, Train/Test Loss: {avg_train_loss:.2f}/{avg_test_loss:.2f}, Policy Rewards {sum(rewards).item():.1f}, Policy Expected Rewards {sum(ER).item():.1f}')

        final_train_loss = avg_train_loss
        final_test_loss = avg_test_loss

        _ = plt.figure(figsize=(10, 7), dpi=400)
        plt.plot(range(1, epochs+1), avg_train_loss_plot, label='train loss')
        plt.plot(range(1, epochs+1), avg_test_loss_plot, label='test loss')
        plt.title('Train and Test Loss During Training')
        plt.xlabel('Epoch')
        plt.ylabel('Mean Squared Error')
        plt.legend()
        plt.savefig(f'./outputs/run{opt_id}/plots/LossPlot.png')
        plt.close()

        fig, axs = plt.subplots(1, 2, figsize=(10, 7), dpi=400)
        axs[0].plot(range(1, epochs+1), torch.flatten(torch.tensor(ER_plot)).detach().numpy(), 'b', label='expected rewards')
        axs[0].legend()

        axs[1].plot(range(1, epochs+1), torch.flatten(torch.tensor(rewards_plot)).detach().numpy(), 'r', label='total rewards')
        axs[1].legend()
        fig.suptitle('Rewards', fontsize=16)
        axs[0].set_xlabel('Epoch')
        axs[1].set_xlabel('Epoch')
        axs[0].set_ylabel('Reward')
        axs[1].set_ylabel('Reward')

        plt.savefig(f'./outputs/run{opt_id}/plots/RewardsPlot.png')
        plt.close()

        final_params = model.transform(model.params).detach().numpy()
        final_cooling = model.transform(model.cooling).detach().numpy()

        return np.concatenate(([opt_id], final_params, final_cooling, [final_train_loss, final_test_loss]))

    def worker_no_cooling(self, opt_id, epochs=100):
        torch.set_num_threads(1)

        # csv_path = '../../Data/DummyData/train5d_sorted.csv'
        # csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/test2d_sorted.csv'

        dir_path = f'./outputs/run{opt_id}/models/'  # where to save models

        policy = None
        model = initialise_model(policy, self.scaling, self.weather_data_path)
        model.save(0, dir_path)  # save initial model

        dt = 30
        sample_size = 24 * 60 ** 2 / dt

        op = OptimiseRC(model, self.csv_path, sample_size, dt, lr=1e-2)

        avg_train_loss_plot = []
        avg_test_loss_plot = []

        plot_results_data = tools.BuildingTemperatureDataset(self.csv_path, 5 * sample_size, all=True)
        plot_dataloader = torch.utils.data.DataLoader(plot_results_data, batch_size=1, shuffle=False)

        # check if dir exists and make if needed
        Path(f'./outputs/run{opt_id}/plots/results/').mkdir(parents=True, exist_ok=True)

        for epoch in trange(epochs):
            # Physical Model Training:
            avg_train_loss = op.train()
            avg_train_loss_plot.append(avg_train_loss)

            # Test Loss
            avg_test_loss = op.test()
            avg_test_loss_plot.append(avg_test_loss)

            # Save a plot of results every 10 epochs or if first/final epoch
            if (epoch + 1) % 10 == 0 or epoch == epochs - 1 or epoch == 0:
                pltsolution_1rm(model, plot_dataloader,
                                f'./outputs/run{opt_id}/plots/results/Result_Epoch{epoch + 1}.png')

            # Save Model
            model_id = epoch + 1
            model.save(model_id, dir_path)

            tqdm.write(f'Epoch {epoch + 1}, Train/Test Loss: {avg_train_loss:.2f}/{avg_test_loss:.2f}')

        final_train_loss = avg_train_loss
        final_test_loss = avg_test_loss

        _ = plt.figure(figsize=(10, 7), dpi=400)
        plt.plot(range(1, epochs + 1), avg_train_loss_plot, label='train loss')
        plt.plot(range(1, epochs + 1), avg_test_loss_plot, label='test loss')
        plt.title('Train and Test Loss During Training')
        plt.xlabel('Epoch')
        plt.ylabel('Mean Squared Error')
        plt.legend()
        plt.savefig(f'./outputs/run{opt_id}/plots/LossPlot.png')
        plt.close()

        final_params = model.transform(model.params).detach().numpy()
        final_cooling = model.transform(model.cooling).detach().numpy()

        return np.concatenate(([opt_id], final_params, final_cooling, [final_train_loss, final_test_loss]))


if __name__ == '__main__':
    use_ray = False

    if use_ray:
        num_cpus = 4
        num_jobs = num_cpus
        epochs = 2

        ray.init(num_cpus=num_cpus)

        # Initialise scaling class
        rm_CA = [100, 1e4]  # [min, max] Capacitance/area
        ex_C = [1e3, 1e8]  # Capacitance
        R = [0.1, 5]  # Resistance ((K.m^2)/W)
        Q_limit = [-10000, 10000]  # Cooling limit in W and gain limit in W/m2
        scaling = InputScaling(rm_CA, ex_C, R, Q_limit)
        weather_data_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/Met Office Weather Files/JuneSept.csv'
        csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/test2d_sorted.csv'

        RayActor = ray.remote(RayActor)

        actors = [RayActor.remote(scaling, weather_data_path, csv_path) for _ in range(num_jobs)]

        results = ray.get([a.worker.remote(num, epochs) for num, a in enumerate(actors)])
        # results = ray.get([a.worker_no_cooling.remote(num, epochs) for num, a in enumerate(actors)])

        time.sleep(3)  # Print is cut short without sleep

        ray.shutdown()

        params_heading = ['Rm Cap/m2 (J/K.m2)', 'Ext Wl Cap 1 (J/K)', 'Ext Wl Cap 2 (J/K)', 'Ext Wl Res 1 (K.m2/W)', 'Ext Wl Res 2 (K.m2/W)', 'Ext Wl Res 3 (K.m2/W)', 'Int Wl Res (K.m2/W)']
        cooling_heading = ['Cooling (W)']
        headings = [['Run Number'], params_heading, cooling_heading, ['Final Average Train Loss', 'Final Avg Test Loss']]
        flat_list = [item for sublist in headings for item in sublist]

        df = pd.DataFrame(np.array(results), columns=flat_list)
        df.to_csv('./outputs/results.csv', index=False,)

    else:
        # Initialise scaling class
        rm_CA = [100, 1e4]  # [min, max] Capacitance/area
        ex_C = [1e3, 1e8]  # Capacitance
        R = [0.1, 5]  # Resistance ((K.m^2)/W)
        Q_limit = [-10000, 10000]  # Cooling limit in W and gain limit in W/m2
        scaling = InputScaling(rm_CA, ex_C, R, Q_limit)
        weather_data_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/Met Office Weather Files/JuneSept.csv'
        csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/test2d_sorted.csv'

        actor = RayActor(scaling, weather_data_path, csv_path)

        results = actor.worker(0, 5)

        params_heading = ['Rm Cap/m2 (J/K.m2)', 'Ext Wl Cap 1 (J/K)', 'Ext Wl Cap 2 (J/K)', 'Ext Wl Res 1 (K.m2/W)',
                          'Ext Wl Res 2 (K.m2/W)', 'Ext Wl Res 3 (K.m2/W)', 'Int Wl Res (K.m2/W)', 'Offset Gain (W/m2)']
        cooling_heading = ['Cooling (W)']
        headings = [['Run Number'], params_heading, cooling_heading,
                    ['Final Average Train Loss', 'Final Avg Test Loss']]
        flat_list = [item for sublist in headings for item in sublist]

        df = pd.DataFrame(np.array([results]), columns=flat_list)
        df.to_csv('./outputs/results.csv', index=False, )





import torch
import numpy as np
from torch import nn
from torchdiffeq import odeint
from datetime import datetime


class RCModel(nn.Module):
    """
    Custom Pytorch model for gradient optimization.
    Initialises with random parameters.

    scaling - Class containing methods to scale inputs from 0-1 back to their usual values and back again.
    transform - function to transform parameters e.g. sigmoid

    Runs using forward method should be sequential as the output is used as the initial condition for the next run, unless manually reset using self.iv
    """

    def __init__(self, building, scaling, Tout_continuous, transform=None, cooling_policy=None):

        super().__init__()
        self.building = building

        self.transform = transform  # transform performed on parameters e.g. sigmoid

        self.scaling = scaling  # InputScaling class (helper class to go between machine (0-1) and physical values)

        self.Tout_continuous = Tout_continuous  # Interp1D object

        self.cooling_policy = cooling_policy  # Neural net: pi(state) --> action

        self.params = None  # initialised in init_params()
        self.cooling = None
        self.init_params()  # initialize weights with random numbers, length of params given from building class

        self.ode_t = None  # Keeps track of t during integration. None is just to initialise attribute
        self.record_action = None  # records Q and t during integration
        self.heat = None  # Heat in Watts

        self.reset_iv()  # sets iv

    def forward(self, t_eval):
        """
        Integrates the ode forward in time.

        building - Initialised RCModel Class
        Tout_continuous - A scipy interp1d function covering the whole of t_eval. Tout_continuous(t) = Outside temperature at time t
        iv - Initial value. Starting temperatures of all nodes
        t_eval - times function should return a solution. e.g. torch.arrange(0, 10000, 30). ensure dtype=float32
        t0 - starting time if not 0
        """
        self.ode_t = -30  # Keeps track of t during ODE integration. Needed to assign log probs only when a step occurs
        self.record_action = []  # Keeps track of action at time during ODE integration.

        # check if t_eval is formatted correctly:
        if t_eval.dim() > 1:
            t_eval = t_eval.squeeze(0)

        # Transform parameters
        if self.transform:
            theta = self.transform(self.params)
            self.heat = self.transform(self.cooling)  # Watts for each room
        else:
            theta = self.params
            self.heat = self.cooling

        # Scale inputs up to their physical values
        theta = self.scaling.physical_param_scaling(theta)
        self.heat = self.scaling.physical_cooling_scaling(self.heat)

        # t_eval_plus = torch.cat((t_eval, (t_eval[-1] + 600).unsqueeze(0)),
        #                         0)  # added on extra time to avoid error with interp1d(t_end)
        # Qrms_continuous = self.building.Q_continuous(t_eval_plus, Q_avg, self.heating[:, 1], self.heating[:, 2])

        # t0 stores the starting epoch time and t_eval is array of seconds from start, [0, 1*dt, 2*dt, ...]
        self.t0 = t_eval[0]
        t_eval = t_eval - self.t0

        # Produce matrix A and B from current parameters
        self.A = self.building.update_inputs(theta)
        self.B = self.building.input_matrix()

        # integrate using fixed step (rk4) see torchdiffeq docs for more options.
        integrate = odeint(self.fODE, self.iv, t_eval, method='rk4')  # https://github.com/rtqichen/torchdiffeq

        return integrate  # first two columns are external envelope nodes. i.e not rooms

    def fODE(self, t, x):
        """
        Provides the function:
        dy/dx = Ax + Bu
        """

        mu = 23.359
        std_dev = 1.41

        x_norm = (x - mu) / std_dev
        x_norm = torch.reshape(x_norm, (-1,))

        day = 24 * 60 ** 2
        week = 7 * day
        # year = (365.2425) * day

        time_signals = [np.sin((t+self.t0) * (2 * np.pi / day)),
                        np.cos((t+self.t0) * (2 * np.pi / day)),
                        np.sin((t+self.t0) * (2 * np.pi / week)),
                        np.cos((t+self.t0) * (2 * np.pi / week))
                        ]

        state_cool = torch.cat((x_norm, torch.tensor(time_signals, dtype=torch.float32)))

        # get cooling action if policy is not None and 15 minutes has passed since last action
        if self.cooling_policy and t - self.ode_t >= 60*15:
            action, log_prob = self.cooling_policy.get_action(state_cool)
            self.ode_t = t
            self.record_action.append([t, action])  # This is just used for plotting the cooling after.

            if self.cooling_policy.training:  # if in training mode store log_prob
                self.cooling_policy.log_probs.append(log_prob)

        else:  # No cooling
            action = 0

        Q = -self.heat * action

        # Q + gain
        Q = Q + self.building.proportional_heating(self.building.gain)

        Tout = self.Tout_continuous(t.item() + self.t0)

        u = self.building.input_vector(Tout, Q)

        return self.A @ x + self.B @ u

    def reset_iv(self):
        # Starting Temperatures of nodes. Column vector of shape ([n,1]) n=rooms+2
        self.iv = 26 * torch.ones(2 + len(self.building.rooms))  # set iv at 26 degrees
        # Set iv as column vector. Errors caused if Row vector which are difficult to trace.
        self.iv = self.iv.unsqueeze(1)

    def init_params(self):
        params = torch.rand(self.building.n_params, dtype=torch.float32, requires_grad=True)

        # enables spread of initial parameters. Otherwise sigmoid(rand) tends towards 0.5.
        if self.transform == torch.sigmoid:
            params = torch.logit(params)  # inverse sigmoid

        # make theta torch parameters
        self.params = nn.Parameter(params)
        self.cooling = nn.Parameter(torch.rand((len(self.building.rooms)), dtype=torch.float32, requires_grad=True))  # [Q]

    def save(self, model_id=0, dir_path=None):
        """
        Save the model, but first transform parameters back to their physical values.
        This makes passing between models with different scaling limits possible.
        Use load method to load back in. Parameters will be converted back to 0-1 using the current scaling function.
        """
        scaled_state_dict = self.state_dict()
        scaled_state_dict['params'] = self.scaling.physical_param_scaling(self.transform(scaled_state_dict['params']))
        scaled_state_dict['cooling'] = self.scaling.physical_cooling_scaling(self.transform(scaled_state_dict['cooling']))

        if dir_path is None:
            dir_path = "./outputs/models/"

        from pathlib import Path
        Path(dir_path).mkdir(parents=True, exist_ok=True)
        torch.save(scaled_state_dict, dir_path + "/rcmodel" + str(model_id) + ".pt")

    def load(self, path):
        """
        Load in model and convert parameters from their physical values to 0-1 using scaling methods.
        Only use if model has been saved with self.save() method.
        """
        scaled_state_dict = torch.load(path)

        # plus 1e-15 used to stop log(0)=-inf
        scaled_state_dict['params'] = torch.logit(
            self.scaling.model_param_scaling(scaled_state_dict['params']) + 1e-15
        )

        scaled_state_dict['cooling'] = torch.logit(
            self.scaling.model_cooling_scaling(scaled_state_dict['cooling']) + 1e-15
        )

        self.load_state_dict(scaled_state_dict)



import numpy as np
import torch
from xitorch.interpolate import Interp1D


class Building:
    """
    A class containing all the rooms of a building (2D)
    """

    def __init__(self, rooms, height=1):

        self.rooms = rooms  # list of Room classes within the building
        self.height = height  # height of all Rooms in the building

        #  Parameters to be properly initialised later in self.update_inputs()
        self.Re = [0, 0, 0]      # external resistance
        self.Ce = [0, 0]      # external capacitance
        self.Rint = 0    # internal wall resistance
        self.gain = 0    # constant heat gain (W/m^2)

        self.Walls = self.sort_walls()


        # get total external area:
        surf_area = 0
        for i in range(len(self.Walls)):
            if self.Walls[i].is_external:
                surf_area += self.Walls[i].area

        self.surf_area = surf_area
        self.n_params = self.get_n_params()  # Number of parameters needed for model i.e len(theta)

    def make_connectivity_matrix(self):
        """
        Creates connectivity matrix of the thermal conductivity (W/K) of rooms in building.
        The first row/col is an external connection.
        Thermal conductivity is calculated by K = area/resistance


        Output matrix is in the form:
            [External TA TB ... Tn]
        [External]
        [TA]
        [TB]
        .
        .
        [Tn]

        """

        n = len(self.rooms)

        knect = torch.zeros([n, n])

        # Iterate through each room and check if room shares any walls with other rooms
        for rm_x in range(0, n - 1):
            for rm_y in range(rm_x + 1, n):

                # compare first wall in rm_x with all walls in rm_y then move on to next wall
                for wl_x in range(len(self.rooms[rm_x].walls)):
                    for wl_y in range(len(self.rooms[rm_y].walls)):

                        if self.rooms[rm_x].walls[wl_x] == self.rooms[rm_y].walls[wl_y]:
                            area = self.Walls[self.rooms[rm_x].walls[wl_x]].area
                            resistance = self.Walls[self.rooms[rm_x].walls[wl_x]].resistance

                            # Sum the area to find total shared area to a room
                            knect[rm_x, rm_y] = knect[rm_x, rm_y] + area / resistance
                            knect[rm_y, rm_x] = knect[rm_x, rm_y]  # matrix is symmetric

        # Now rooms are checked if they are external and stored in a vector.
        is_ex = torch.zeros(n)
        for rm in range(n):
            for wl in range(len(self.rooms[rm].walls)):
                wl_class = self.Walls[self.rooms[rm].walls[wl]]

                if wl_class.is_external:
                    is_ex[rm] = is_ex[rm] + wl_class.area / wl_class.resistance

        # The is_external vector is used as the first row/col for the connection matrix
        knect = torch.vstack((is_ex, knect))
        is_ex = torch.cat(
            (torch.tensor([0]), is_ex))  # adds 0 to start of array (account for external to external connection)

        is_ex = is_ex.unsqueeze(0)  # adds dimension
        knect = torch.hstack((is_ex.T, knect))

        return knect

    def make_system_matrix(self):
        """
        Creates the system/state matrix (A) for the Building.
        This will be used in the State Space equation: xdot = A@x + B@u

        The system matrix (A) is in the form:

        x = [Te1, Te2, TA , TB , ... Tn ]

        A = [A00, A01, A02, A03, ... A0n], #dTe1/dt
            [A10, A11, A12, A13, ... A1n], #dTe2/dt
            [A20, A21, A22, A23, ... A2n], #dTA/dt
            .
            .
            .

        """

        knect = self.make_connectivity_matrix()

        n = len(knect[0])

        off = len(self.Ce) - 1  # Num nodes not seen in connection matrix which need to be offset in A.

        # Plus "off" accounts for the node Te1 in the external envelope which is not part of the connection matrix
        A = torch.zeros([n + off, n + off])

        # Hard code in the external to external connections. These will always be the same.
        A[0, 0] = self.surf_area * (-1 / (self.Re[0] * self.Ce[0]) - 1 / (self.Re[1] * self.Ce[0]))
        A[0, 1] = self.surf_area / (self.Re[1] * self.Ce[0])
        A[1, 0] = self.surf_area / (self.Re[1] * self.Ce[1])
        A[1, 1] = -self.surf_area / (self.Re[1] * self.Ce[1])

        for row in range(n):
            if row == 0:
                c = self.Ce[-1]
            else:
                c = self.rooms[row - 1].capacitance  # Heat Capacity (J/K)

            for col in range(n):
                K = knect[row, col]  # Thermal Conductance (W/K) K = A(m^2)/R(K.m^2/W)

                A[row + off, col + off] = A[row + off, col + off] + K / c
                A[row + off, row + off] = A[row + off, row + off] - K / c

        return A

    def sort_walls(self):
        """
        Function which combines 3 separate functions in the correct order.
        All unique walls in the building are represented in a list of Wall class instances.
        Walls previously defined in the Room class are also updated to be an index of this list.
        """

        Walls = self.get_walls()
        self.update_rooms(Walls)
        Walls = self.get_external_walls(Walls)

        return Walls

    def get_walls(self):
        """
        Looks at every wall in every room within the building and produces a list of unique walls.
        This list is then used to create a Wall class for each unique wall.
        Output is list all Wall classes. [Wall1, Wall2, ... Walln]
        """
        # First get list of all walls in all rooms.
        walls = []
        for rm in range(len(self.rooms)):
            for wl in range(len(self.rooms[rm].walls)):
                walls.append(self.rooms[rm].walls[wl])

        # Now the duplicates need to be deleted so only unique walls remain.
        duplicate_wls = []
        for wl1 in range(len(walls) - 1):
            for wl2 in range(wl1 + 1, len(walls)):
                if set(walls[wl1]) == set(walls[wl2]):
                    duplicate_wls.append(wl1)  # A list of indexes is created

        # Delete duplicates from the original list using the indexes found above
        for index in sorted(duplicate_wls, reverse=True):
            del walls[index]

        # A Wall class is instantiated for each unique wall and saved in the list Walls
        Walls = []
        for wl in walls:
            Walls.append(self.Wall(wl, self.Rint, self.height))

        return Walls

    def update_rooms(self, Walls):
        """
        Update the Room.walls variable to be an index matching the list Walls which contains the class instances of each unique wall.
        """
        # We can now use this list, Walls, as an index for the Room classes.
        for rm in range(len(self.rooms)):
            for wl in range(len(self.rooms[rm].walls)):
                for Walls_indx in range(len(Walls)):

                    # For each wall in the room class replace with an index to the matching wall in the list Walls
                    if set(self.rooms[rm].walls[wl]) == set(Walls[Walls_indx].coordinates):
                        self.rooms[rm].walls[wl] = Walls_indx
                        break

                    # Check for no matches
                    elif Walls_indx == len(Walls):
                        print("Error: Room: ", rm, ", Wall: ", wl, ". Was not matched")

    def get_external_walls(self, Walls):
        """
        Finds all external walls and updates the boolean variable is_external in the Wall Class.
        """

        # A wall is external if it is not shared by another room.
        rm_wls = []
        for rm in range(len(self.rooms)):
            for wl in range(len(self.rooms[rm].walls)):
                rm_wls.append(
                    self.rooms[rm].walls[wl])  # Creates list of walls in each room. Can now check for multiples.

        for i in rm_wls:
            occurrences = torch.count_nonzero(torch.tensor(rm_wls) == i)

            # Occurance of 1 means the wall is only seen in a single room meaning it must be an external wall.
            if occurrences == 1:
                Walls[i].is_external = True
                Walls[i].resistance = self.Re[2]  # Change resistance of wall

        return Walls

    def input_vector(self, Tout, Q):
        """
        Input vector in form:
        u = [Tout, QA, QB, ... Qn]
        """
        # Check and make 2d matrix if 1d.
        # add_dim = lambda x: torch.tensor([x]) if x.ndim == 1 else x

        # Q = np.array(Q)
        # Q = add_dim(Q)
        # Q = torch.tensor(Q, dtype=torch.float32).unsqueeze(0) #add dim
        Q = Q.unsqueeze(0)  # add dim

        Tout = Tout.type(torch.float32)
        Tout = torch.reshape(Tout, (1, 1))

        if len(Q[0]) == len(self.rooms):
            u = torch.cat((Tout, Q), dim=1).T

            return u

        else:
            raise ValueError("input_vector: Q needs to have 1 column per room")

    def input_matrix(self):
        """
        Produces the input matrix: B
        """
        # u = [[Tout, QA, QB,...]].T

        off = len(self.Ce)  # offset needed for external nodes
        num_inputs = len(self.rooms) + 1

        B = torch.zeros((off + len(self.rooms), num_inputs))

        # Set Tout input:
        B[0, 0] = self.surf_area / (self.Re[0] * self.Ce[0])

        # Set the Q inputs for each room
        for rm in range(len(self.rooms)):
            B[rm + off, rm + 1] = 1 / self.rooms[rm].capacitance

        return B

    def update_wall(self, wall_ID, resistance=-1):
        """
        Used to update the resistance of the wall.
        Wall is chosen by either supplying the coordinates of the wall or the index number to the Walls list.
        Coordinates must be supplied in a tuple e.g. ((x1, y1), (x2, y2))
        Function returns its wall index number in the list Walls
        """
        is_coord = False

        # check if ID is a coordinate or an index
        if isinstance(wall_ID, int):
            wl_indx = wall_ID
        elif len(self.Walls[0].coordinates) == len(wall_ID):
            is_coord = True
        else:
            return print("wall_ID has been input incorrectly")

        if is_coord:
            for wl in range(len(self.Walls)):

                if set(self.Walls[wl].coordinates) == set(tuple(wall_ID)):
                    wl_indx = wl
                    break
        try:
            if resistance != -1:
                self.Walls[wl_indx].resistance = resistance


        except:
            print("Could not find the wall")

        return wl_indx

    def plt_ext_walls(self):
        """
        Plots all walls and highlights in red the external walls. Used to check if model has correctly identified
        """
        from matplotlib import pyplot as plt
        import numpy as np

        for wl in range(len(self.Walls)):
            c = np.array(self.Walls[wl].coordinates)
            if self.Walls[wl].is_external:
                ex, = plt.plot(c[:, 0], c[:, 1], 'r', label='External Walls')
            else:
                int, = plt.plot(c[:, 0], c[:, 1], 'k', label='Internal Walls')

        plt.legend([ex, int], ['External Walls', 'Internal Walls'])

        return plt.show()

    def categorise_theta(self, theta):
        """
        Function to split theta into its different categories.

        theta is a 1D vector with categories in the following order:
        theta = [room capacitance, external capacitance, external resistance, internal wall resistance, gain]
        """

        indx = 1
        rm_cap = theta[0:indx]
        indx += 2
        ex_cap = theta[indx - 2:indx]
        indx += 3
        ex_r = theta[indx - 3:indx]
        indx += 1
        wl_r = theta[indx - 1]
        indx += 1
        gain = theta[indx - 1]

        return rm_cap, ex_cap, ex_r, wl_r, gain

    def update_inputs(self, theta):
        """
        Function to update the instance with new variables, theta. Must be structured correctly.
        Note: system matrices will have to be reproduced e.g. A, B

        theta is a 1D vector with categories in the following order:
        theta = [room capacitance, external capacitance, external resistance, each wall resistance, gain]


        returns Q_avg: The mean W/m^2 for the building
        """

        rm_cap, ex_cap, ex_r, wl_r, gain = self.categorise_theta(theta)

        # update room capacitance
        for i in range(len(self.rooms)):
            self.rooms[i].capacitance = rm_cap[0] * self.rooms[i].area

        # update external capacitance
        self.Ce = ex_cap

        # update external resistance
        self.Re = ex_r

        # update wall resistance
        self.Rint = wl_r
        # This code is not great, need way of assigning wall resistance individually in the future
        for i in range(len(self.Walls)):
            if self.Walls[i].is_external:
                self.Walls[i].resistance = ex_r[2]
            else:
                self.Walls[i].resistance = wl_r


        # update gain
        self.gain = gain

        A = self.make_system_matrix()

        return A

    def get_n_params(self):
        """Find how many parameters the model needs by testing each number."""
        n_params = 1
        while True:
            params = torch.ones(n_params)
            try:
                self.categorise_theta(params)
            except:
                n_params += 1

            else:
                break  # This is reached if no error occurs

        return n_params

    # Heating:
    def proportional_heating(self, Q_avg):
        """
        Input: Q_avg (W/m^2)
        Calc: Q_avg * rm_area
        Outputs: Q (Watts)
        """

        Q = torch.zeros(len(self.rooms))
        for i in range(len(self.rooms)):
            Q[i] = Q_avg * self.rooms[i].area

        return Q

    def time_angle(self, circ):
        """outputs angle from x axis to point on circle. Used to represent time in a day"""

        theta = np.angle(circ) * 180 / np.pi
        # add 360 if y axis is negative.
        theta = np.where(circ.imag < 0, theta + 360, theta)

        return torch.tensor(theta)

    def Q_control(self, t, bound_A, bound_B):
        """
        Produces a step function for given timeseries, turn on when time of day is between the thresholds A-B.
        t is unix epoch time (seconds since 00:00 Jan 1970), t is converted to an angle (degrees 0-360) repeating every 24hrs.
        Theta_A and Theta_B represent the start and stop day angle.


        t - time evaluations, unix epoch time
        bound_* - enter value between -6:6 this is then scaled up to between 0 and 360 using a sigmoid function.

        Output: step function off (0) or on (1) for each point of the timeseries. Each row is a different room.
                tensor([[rm1],
                        [rm2],
                        ....
                        [rmn]])

        """

        s_per_day = 24 * 60 ** 2

        # function which bounds value between 0:360
        bound = lambda x: 360 * torch.sigmoid(x)

        # scale theta between 0:360 degrees
        theta_A = bound(bound_A)
        theta_B = bound(bound_B)

        # get angle (i.e time of day) for timeseries
        circ = np.e ** (1j * 2 * np.pi / s_per_day * t)
        time_degree = self.time_angle(circ.cpu())

        # Control logic:
        # turn Q on if time_degree is between thresholds theta_A and theta_B.
        # Do this for each room. num rooms = len(theta_A)

        Q_on_off = torch.zeros((len(theta_A), len(t)))
        stretch = 2e1  # the larger this number the steeper the step function is. But you lose the ability to autograd the gradient.
        # 2e1 means the step goes from 0-1 in roughly two minutes and grad is calculated fine.
        for i in range(len(theta_A)):
            if theta_A[i] < theta_B[i]:
                condition1 = torch.sigmoid((time_degree - theta_A[i]) * stretch)  # True if time>A
                condition2 = torch.sigmoid((theta_B[i] - time_degree) * stretch)  # True if time<B
                Q_on_off[i] = torch.sigmoid(
                    (condition1 + condition2 - 1.9) * stretch)  # True if con1 & con2 on between A-B
            else:
                condition1 = torch.sigmoid((time_degree - theta_A[i]) * stretch)
                condition2 = torch.sigmoid((theta_B[i] - time_degree) * stretch)
                Q_on_off[i] = torch.sigmoid((condition1 + condition2 - 0.4) * stretch)  # on between B-A

        return Q_on_off

    def Q_continuous(self, t, Q_avg, theta_A1, theta_B1):

        Q_on_off = self.Q_control(t, theta_A1, theta_B1)  # get control step funciton for each room

        Q_on_off = Q_on_off

        Q = Q_on_off * Q_avg.unsqueeze(1)  # multiply by Q_avg for each room

        Q_cont = Interp1D(t, Q, method='linear')

        return Q_cont

    class Wall:
        """
        A nested class, this is used to contain unique information about each wall in the building.
        """

        def __init__(self, coordinates, resistance, height):
            self.coordinates = coordinates
            self.resistance = resistance
            self.is_external = False  # This is updated later
            self.height = height
            self.area = self.get_area()

        def get_area(self):
            wl = torch.tensor(self.coordinates)  # put wall in array for easy calc of area
            length = torch.linalg.norm(wl[0] - wl[1], dtype=torch.float)

            return length * self.height



from .building import Building
import torch
from torch.utils.data import Dataset
import pandas as pd
import matplotlib.pyplot as plt


class InputScaling(Building):
    """
    Class to group methods for scaling input parameters.
    Provide the limits for each variable. e.g. rm_CA = [100, 1000]

    Model scaling - scales physical parameters to between 0-1 using min max.
    Physical scaling - returns parameters back to their physical meaning.

    Initialise with:
    InputScaling(rm_CA ex_C, R, Q_limit)
    or
    InputScaling(input_range=input_range)
    """

    def __init__(self, rm_CA=None, ex_C=None, R=None, Q_limit=None, input_range=None):

        if input_range is None:
            input_range = [rm_CA, ex_C, R, Q_limit]

            # check for None
            if None in input_range:
                raise ValueError('None type in list. Define all inputs or use a predefined input_range instead')

        self.input_range = torch.tensor(input_range)

    def physical_param_scaling(self, theta_scaled):
        """
        Scale from 0-1 back to normal.
        """

        rm_cap, ex_cap, ex_r, wl_r, gain = self.categorise_theta(theta_scaled)

        rm_cap = self.unminmaxscale(rm_cap, self.input_range[0])
        ex_cap = self.unminmaxscale(ex_cap, self.input_range[1])
        ex_r   = self.unminmaxscale(ex_r,   self.input_range[2])
        wl_r   = self.unminmaxscale(wl_r,   self.input_range[2])
        gain   = self.unminmaxscale(gain,   self.input_range[3])

        if wl_r.ndim == 0:
            wl_r = torch.unsqueeze(wl_r, 0)
        if gain.ndim == 0:
            gain = torch.unsqueeze(gain, 0)

        theta = torch.cat([rm_cap, ex_cap, ex_r, wl_r, gain])

        return theta

    def model_param_scaling(self, theta):
        """
        Scale to 0-1.
        """

        rm_cap, ex_cap, ex_r, wl_r, gain = self.categorise_theta(theta)

        rm_cap = self.minmaxscale(rm_cap, self.input_range[0])
        ex_cap = self.minmaxscale(ex_cap, self.input_range[1])
        ex_r   = self.minmaxscale(ex_r  , self.input_range[2])
        wl_r   = self.minmaxscale(wl_r  , self.input_range[2])
        gain   = self.minmaxscale(gain  , self.input_range[3])

        if wl_r.ndim == 0:
            wl_r = torch.unsqueeze(wl_r, 0)
        if gain.ndim == 0:
            gain = torch.unsqueeze(gain, 0)

        theta_scaled = torch.cat([rm_cap, ex_cap, ex_r, wl_r, gain])

        return theta_scaled

    def physical_cooling_scaling(self, Q):

        Q_watts = self.unminmaxscale(Q, [0, self.input_range[3].max()])

        return Q_watts

    def model_cooling_scaling(self, Q_watts):

        Q = self.minmaxscale(Q_watts, [0, self.input_range[3].max()])

        return Q

    def minmaxscale(self, x, x_range):
        if not torch.is_tensor(x_range):
            x_range = torch.tensor(x_range)
        if not torch.is_tensor(x):
            x = torch.tensor(x)

        x_scaled = (x - x_range.min()) / (x_range.max() - x_range.min())

        return x_scaled

    def unminmaxscale(self, x_scaled, x_range):
        if not torch.is_tensor(x_range):
            x_range = torch.tensor(x_range)
        if not torch.is_tensor(x_scaled):
            x_scaled = torch.tensor(x_scaled)

        x = x_scaled * (x_range.max() - x_range.min()) + x_range.min()

        return x


class BuildingTemperatureDataset(Dataset):
    """
    Splits dataset up into batches of len(dataset) // sample_size. Note remainder of data is thrown away.
    train and test tags can be used to select a percentage slice of data in this order, see function _split_dataset() for current % splits.

    If there is insufficient data for one batch, sample_size will be reduced to match the data.
    """
    def __init__(self, csv_path, sample_size, transform=None, all=True, train=False, test=False):
        self.csv_path = csv_path
        self.transform = transform
        self.sample_size = int(sample_size)
        self.headings = list(pd.read_csv(csv_path, nrows=1)) # list of dataframe headings
        self.all = all
        self.train = train
        self.test = test

        # auto splits data by train and test
        # entry count is number of rows to read from csv.
        # remainder of data e.g. after entry_count//sample_size is lost
        self.rows_to_skip, self.entry_count = self._split_dataset()

        self.__len__()  # Used to initialise logic used in __len__.

    def __len__(self):
        """
        Get number of batches in the dataset. Returns int
        Minimum of 1 batch will be returned.
        """

        num_samples = self.entry_count//self.sample_size

        # Insufficient data for 1 whole sample size. Remainder of data used instead.
        if num_samples == 0:
            num_samples = 1
            self.sample_size = self.entry_count  # Reduce sample to the entries remaining

        return num_samples

    def __getitem__(self, idx):
        """
        Returns a 'sample' of the total dataset.

        Each sample has length 'sample_size' except for the final one, which contains the remainder of the dataset.

        The entire dataset is never loaded in at once, each sample is read in separately when needed to reduce memory usage.
        """
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # Get upper and lower bounds of dataset slice
        lb = idx*self.sample_size + self.rows_to_skip
        ub = (idx+1)*self.sample_size + self.rows_to_skip

        # Get pandas df of sample
        df_sample = pd.read_csv(self.csv_path, skiprows=lb, nrows=self.sample_size)

        # Get time column (time must be in the 1th column)
        t_sample = torch.tensor(df_sample.iloc[:, 1].values, dtype=torch.float64) # units (s)

        # Get temp matrix
        temp_sample = torch.tensor(df_sample.iloc[:, 2:].values, dtype=torch.float32) #pandas needs 2: to get all but first & second column

        #apply transorms if required
        if self.transform:
            temp_sample = self.transform(temp_sample)

        return t_sample, temp_sample

    def _get_entries(self):
        """Get total rows/entries of data in csv"""
        from csv import reader

        # the rows in the .csv are counted:
        with open(self.csv_path,"r") as f:
            read_f = reader(f,delimiter = ",")
            entry_count = sum(1 for row in read_f) - 1 # minus one to account for heading

        return entry_count

    def _split_dataset(self):

        train_split = 0.8
        test_split = 0.2

        total_entries = self._get_entries() #total rows of data in csv

        if self.train:
            rows_to_skip = 0
            entry_count = int(total_entries*train_split)
            return rows_to_skip, entry_count
        elif self.test:
            rows_to_skip = int(total_entries*train_split)
            entry_count = int(total_entries*test_split)
            return rows_to_skip, entry_count
        elif self.all:
            rows_to_skip = 0
            entry_count = total_entries
            return rows_to_skip, entry_count
        else:
            raise ValueError('train, test and validation all False')


# helper functions:
def pltsolution_1rm(model, dataloader, filename=None):
    """
    Plots the first sample of dataloader.
    """
    model.reset_iv()  # Reset initial value
    model.eval()  # Put model in evaluation mode

    # Get solution ---------------
    time, data_temp = next(iter(dataloader))

    time = time.squeeze(0)
    data_temp = data_temp.squeeze(0)

    t_days = (time - time.min()) / (24 * 60 ** 2)  # Time in days

    pred = model(time)
    pred = pred.squeeze(-1)

    # Get Heating Control for each room
    if model.cooling_policy:
        record_action = torch.tensor(model.record_action)
        Q_tdays = record_action[:, 0] / (24 * 60 ** 2)  # Time in days
        Q_on_off = record_action[:, 1:]  # Cooling actions

        Q_avg = model.transform(model.cooling)
        Q_avg = model.scaling.physical_cooling_scaling(Q_avg)

        Q = Q_on_off * -Q_avg.unsqueeze(1)  # Q is timeseries of Watts for each room.
    else:
        Q_tdays = t_days
        Q = torch.zeros(len(Q_tdays))

    # Compute and print loss.
    # loss_fn = torch.nn.MSELoss()
    # num_cols = len(model.building.rooms)  # number of columns to take from data.
    # loss = loss_fn(pred[:, 2:], data_temp[:, 0:num_cols])

    # print(f"Test loss = {loss.item():>8f}")

    # ---------------------------------

    # Plot Solution

    ax2ylim = 250

    fig, axs = plt.subplots(figsize=(10, 8))

    ax2 = axs.twinx()
    ln1 = axs.plot(t_days.detach().numpy(), pred[:, 2:].detach().numpy(), label='model')
    ln2 = axs.plot(t_days.detach().numpy(), data_temp[:, 0].detach().numpy(), label='data')
    ln3 = axs.plot(t_days.detach().numpy(), model.Tout_continuous(time).detach().numpy(), label='outside')
    ln4 = ax2.plot(Q_tdays.detach().numpy(), Q.detach().numpy(), '--', color='black', alpha=0.5, label='heat')
    axs.set_title(model.building.rooms[0].name)
    ax2.set_ylabel(r"Heating/Cooling ($W/m^2$)")
    # ax2.set_ylim(-ax2ylim, ax2ylim)

    lns = ln1 + ln2 + ln3 + ln4
    labs = [l.get_label() for l in lns]
    axs.legend(lns, labs, loc=0)

    axs.set(xlabel='Time (days)', ylabel='Temperature ($^\circ$C)')

    if filename:
        fig.savefig(filename)
        plt.close()

    else:
        plt.show()


import pandas as pd
import matplotlib.pyplot as plt
import torch
import numpy as np
from pathlib import Path
from tqdm.auto import tqdm, trange

from rcmodel import *
from main import initialise_model, RayActor

# Laptop
weather_data_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/Met Office Weather Files/JuneSept.csv'
csv_path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/train5d_sorted.csv'

# Hydra:
# weather_data_path = '/home/benf/LSI/Data/Met Office Weather Files/JuneSept.csv'
# csv_path = '/home/benf/LSI/Data/DummyData/train5d_sorted.csv'  # where building data

opt_id = 0
dir_path = f'./outputs/run{opt_id}/models/'  # where to save
# Load model?
# load_model_path = 'rcmodel_start.pt'  # or None
load_model_path = None

def do_plots():
    fig = plt.figure(figsize=(10, 7), dpi=400)
    plt.plot(range(1, len(avg_train_loss_plot) + 1), avg_train_loss_plot, label='train loss')
    plt.plot(range(1, len(avg_test_loss_plot) + 1), avg_test_loss_plot, label='test loss')
    plt.title('Train and Test Loss During Training')
    plt.xlabel('Epoch')
    plt.ylabel('Mean Squared Error')
    plt.legend()
    fig.savefig(f'./outputs/run{opt_id}/plots/LossPlot.png')
    plt.close()

    fig, axs = plt.subplots(1, 2, figsize=(10, 7), dpi=400)
    y = torch.flatten(torch.tensor(ER_plot)).detach().numpy()
    axs[0].plot(range(1, len(y) + 1), y, 'b', label='expected rewards')
    axs[0].legend()

    y = torch.flatten(torch.tensor(rewards_plot)).detach().numpy()
    axs[1].plot(range(1, len(y) + 1), y, 'r', label='total rewards')
    axs[1].legend()
    fig.suptitle('Rewards', fontsize=16)
    axs[0].set_xlabel('Epoch')
    axs[1].set_xlabel('Epoch')
    axs[0].set_ylabel('Reward')
    axs[1].set_ylabel('Reward')

    fig.savefig(f'./outputs/run{opt_id}/plots/RewardsPlot.png')
    plt.close()


def init_scaling():
    # Initialise scaling class
    rm_CA = [100, 1e4]  # [min, max] Capacitance/area
    ex_C = [1e3, 1e8]  # Capacitance
    R = [0.1, 5]  # Resistance ((K.m^2)/W)
    scaling = InputScaling(rm_CA, ex_C, R)
    return scaling


# Initialise Model
scaling = init_scaling()
policy = PolicyNetwork(7, 2)
model = initialise_model(policy, scaling, weather_data_path)

model.save(0, dir_path)  # save initial model

# Initialise Optimise Class - for training physical model
dt = 30
sample_size = 24 * 60 ** 2 / dt  # ONE DAY
op = OptimiseRC(model, csv_path, sample_size, dt, lr=1e-3, opt_id=opt_id)

# Initialise Reinforcement Class - for training policy
time_data = torch.tensor(pd.read_csv(csv_path, skiprows=0).iloc[:, 1], dtype=torch.float64)
temp_data = torch.tensor(pd.read_csv(csv_path, skiprows=0).iloc[:, 2:].to_numpy(dtype=np.float32), dtype=torch.float32)
env = LSIEnv(model, time_data)
rl = Reinforce(env, time_data, temp_data, alpha=1e-2)

# lists to keep track of process, used for plot at the end
avg_train_loss_plot = []
avg_test_loss_plot = []
rewards_plot = []
ER_plot = []
count = 1  # Counts number of epochs since start.

# Dataloader to be used when plotting a model run. This is just for info.
plot_results_data = tools.BuildingTemperatureDataset(csv_path, 5 * sample_size, all=True)
plot_dataloader = torch.utils.data.DataLoader(plot_results_data, batch_size=1, shuffle=False)

# check if dir exists and make if needed
Path(f'./outputs/run{opt_id}/plots/results/').mkdir(parents=True, exist_ok=True)


if load_model_path:
    model.load(load_model_path)


start_num = 0  # Number of cycles to start at. Used if resuming the run. i.e. the first cycle is (start_num + 1)

# Convergence happens when the mean gradient of loss/reward is < tol.
tol_phys = 0.01  # 1%
tol_policy = 0.02  # 2%
window_len = 10.  # Length of convergence look-back window.

cycles = 50
max_epochs = 200

plt.ioff()  # Reduces memory usage by matplotlib
for cycle in trange(cycles):

    # -------- Physical model training --------
    tqdm.write('Physical Model Training:')
    r_prev = (torch.arange(window_len) + 1) * 10  # dummy rewards to prevent convergence on first epochs
    loss_prev = 0
    for epoch in range(max_epochs):
        avg_train_loss = op.train()
        avg_train_loss_plot.append(avg_train_loss)

        # Get difference from previous
        indx = epoch % len(r_prev)  # cycles the indexes in array
        r_prev[indx] = avg_train_loss  # keeps track of a window of differences


        # Test Loss
        avg_test_loss = op.test()
        avg_test_loss_plot.append(avg_test_loss)

        tqdm.write(
            f'Epoch {count}, Train/Test Loss: {avg_train_loss:.2f}/{avg_test_loss:.2f}, Mean diff: {abs((r_prev.mean() - r_prev[indx]) / r_prev[indx]) * 100:.3f} %')

        # Save Model
        model_id = count + start_num
        model.save(model_id, dir_path)
        count += 1

        # check if percentage change from mean is less than tol. i.e. convergence
        if abs((r_prev.mean() - r_prev[indx]) / r_prev[indx]) < tol_phys:
            break

    tqdm.write(f'Physical converged in {epoch + 1} epochs. Total epochs: {count - 1}\n')

    # Save a plot of results after physical training
    pltsolution_1rm(model, plot_dataloader,
                    f'./outputs/run{opt_id}/plots/results/Result_Cycle{start_num + cycle + 1}a.png')

    # -------- Policy training --------
    tqdm.write('Policy Training:')
    r_prev = (torch.arange(window_len) + 1) * 1000
    for epoch in range(max_epochs):
        rewards, ER = rl.train(1, sample_size)

        indx = epoch % len(r_prev)
        r_prev[indx] = torch.tensor(rewards)  # keeps track of a window of differences
        rewards_plot.append(rewards)
        ER_plot.append(ER)

        tqdm.write(
            f'Epoch {count}, Rewards/Expected Rewards: {torch.tensor(rewards).sum().item():.2f}/{torch.tensor(ER).sum().item():.2f}, Mean diff: {abs((r_prev.mean() - r_prev[indx]) / r_prev[indx]) * 100:.3f} %')

        # Save Model
        model_id = count + start_num
        model.save(model_id, dir_path)
        count += 1

        # check if percentage change from mean is less than tol. i.e. convergence
        if abs((torch.mean(r_prev) - r_prev[indx]) / r_prev[indx]) < tol_policy:
            break

    tqdm.write(f'Policy converged in {epoch + 1} epochs. Total epochs: {count - 1}\n')

    # Save a plot of results after policy training
    pltsolution_1rm(model, plot_dataloader, f'./outputs/run{opt_id}/plots/results/Result_Cycle{start_num + cycle + 1}b.png')

    # Plot loss and reward plot
    do_plots()





import gym
from gym import spaces

import torch
import torch.nn as nn
from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
from xitorch.interpolate import Interp1D
from tqdm.auto import tqdm, trange
import time

from rcmodel.room import Room
from rcmodel.building import Building
from rcmodel.RCModel import RCModel
from rcmodel.tools import InputScaling
from rcmodel.tools import BuildingTemperatureDataset


class PolicyNetwork(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        n = 10
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(in_dim, n),
            nn.ReLU(),
            nn.Linear(n, n),
            nn.ReLU(),
            nn.Linear(n, out_dim),
        )

        self.on_policy_reset()

    def forward(self, state):
        logits = self.linear_relu_stack(state)

        return logits

    def get_action(self, state):
        pd = torch.distributions.categorical.Categorical(logits=self.forward(state))  # make a probability distribution

        action = pd.sample()  # sample from distribution pi(a|s) (action given state)

        return action, pd.log_prob(action)

    def on_policy_reset(self):
        # this stores log_probs during an integration step.
        self.log_probs = []
        # self.rewards = []


class Reinforce:
    def __init__(self, env, time_data, temp_data, gamma=0.99, alpha=1e-3):

        assert len(time_data) == len(temp_data)

        self.env = env
        # self.pi = pi
        self.time_data = time_data
        self.temp_data = temp_data
        self.gamma = gamma
        self.alpha = alpha

        self.optimiser = torch.optim.Adam(self.env.RC.cooling_policy.parameters(), lr=self.alpha)

    def update_policy(self, rewards, log_probs):
        # downsample rewards by averaging each window.
        rewards = torch.tensor([window.mean() for window in torch.tensor_split(rewards, len(log_probs))])

        # Calculate Discounted Reward:
        discounted_rewards = torch.zeros(len(rewards))

        R = 0
        indx = len(rewards) - 1
        for r in reversed(rewards):
            R = r + self.gamma * R  # Discounted Reward is calculated from last reward to first.

            discounted_rewards[indx] = R  # Fill array back to front to un-reverse the order
            indx -= 1

        # Normalise rewards
        discounted_rewards = (discounted_rewards - discounted_rewards.mean()) / (
                discounted_rewards.std() + 1e-9)

        # discounted_rewards = torch.tensor(np.array(discounted_rewards.detach().numpy()))

        expected_reward = -torch.stack(log_probs) * discounted_rewards  # negative for maximising
        expected_reward = torch.sum(expected_reward)
        # print(f'ER {expected_reward}')

        # Update parameters in pi
        self.optimiser.zero_grad()
        expected_reward.backward()
        self.optimiser.step()

        # print(list(self.pi.parameters())[0].grad)  # check on grads if needed

        return expected_reward

    def train(self, num_episodes, step_size):
        self.env.RC.cooling_policy.train()  # Put in training mode
        total_ER = []
        total_rewards = []

        loss_fn = torch.nn.MSELoss(reduction='none')  # Squared Error

        # with tqdm(total=len(self.env.time_data) * num_episodes, position=0, leave=False) as pbar:  # progress bar
        for episode in range(num_episodes):
            self.env.reset()

            episode_rewards = []
            episode_ER = []

            # Time is increased in steps, with the policy updating after every step.
            while self.env.t_index < len(self.env.time_data) - 1:
                # takes a step_size forward in time
                pred = self.env.step(step_size).squeeze(-1)  # state and action produced in step

                actual = self.temp_data[self.env.t_index:int(self.env.t_index + step_size), 0:self.env.n_rooms]

                # negative so reward can be maximised
                reward = -loss_fn(pred[:, 2:], actual)

                # Do gradient decent on sample
                ER = self.update_policy(reward, self.env.RC.cooling_policy.log_probs)

                self.env.RC.cooling_policy.on_policy_reset()  # empty buffer

                # get last output and use for next initial value
                self.env.RC.iv = pred[-1, :].unsqueeze(1).detach()  # MUST DETACH GRAD

                episode_rewards.append(sum(reward))
                episode_ER.append(ER)

                self.env.t_index += int(step_size)  # increase environment time

            # print(f'Episode {episode+1}, Expected Reward: {sum(episode_ER).item():.2f}, total_reward: {sum(episode_rewards).item():.2f}')

            total_ER.append(sum(episode_ER).detach())
            total_rewards.append(sum(episode_rewards).detach())

        return total_rewards, total_ER


class LSIEnv(gym.Env):
    """Custom Environment that follows gym interface"""
    metadata = {'render.modes': ['human']}

    def __init__(self, RC, time_data):
        super().__init__()

        self.RC = RC  # RCModel Class
        self.time_data = time_data  # timeseries
        self.t_index = 0  # used to keep track of index through timeseries

        # ----- GYM Stuff -----
        self.n_rooms = len(self.RC.building.rooms)
        self.low_state = -10
        self.high_state = 50
        # Define action and observation space
        # They must be gym.spaces objects
        # Example when using discrete actions:
        self.action_space = spaces.Discrete(2, )

        # Observation is temperature of each room.
        self.observation_space = spaces.Box(
            low=self.low_state,
            high=self.high_state,
            shape=(self.n_rooms,),
            dtype=np.float32
        )

    #                  action
    def step(self, step_size):
        # Execute a chunk of timeseries
        t_eval = self.time_data[self.t_index:int(self.t_index + step_size)]

        # actions are decided and stored by the policy while integrating the ODE:
        pred = self.RC(t_eval)

        return pred.detach()  # No need for grad

    #          (observation, reward, done, info)
    # self.state, reward, done, {}

    def reset(self):
        # Reset the state of the environment to an initial state
        self.t_index = 0
        self.RC.reset_iv()

        self.RC.cooling_policy.on_policy_reset()

    def render(self, mode='human', close=False):
        # Render the environment to the screen

        return

if __name__ == '__main__':
    from main import initialise_model

    path_sorted = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/210813data_sorted.csv'
    time_data = torch.tensor(pd.read_csv(path_sorted, skiprows=0).iloc[:, 1], dtype=torch.float64)
    temp_data = torch.tensor(pd.read_csv(path_sorted, skiprows=0).iloc[:, 2:].to_numpy(dtype=np.float32),
                             dtype=torch.float32)

    ######
    path = '/Users/benfourcin/OneDrive - University of Exeter/PhD/LSI/Data/DummyData/'
    dt = 30  # timestep (seconds), data and the model are sampled at this frequency
    sample_size = int(5 * (60 ** 2 * 24) / dt)  # one day of data

    training_data = BuildingTemperatureDataset(path + 'train5d.csv', sample_size)
    train_dataloader = torch.utils.data.DataLoader(training_data, batch_size=1, shuffle=False)
    ######




    time_data = time_data[0:100]
    temp_data = temp_data[0:100, :]

    policy = PolicyNetwork(7, 2)
    RC, Tout_continuous = initialise_model(policy)
    env = LSIEnv(RC, time_data)
    reinforce = Reinforce(env, time_data, temp_data, alpha=1e-2)

    num_episodes = 10
    step_size = 24*60**2 / 30  # timesteps in 1 day
    start_time = time.time()
    plot_total_rewards, plot_ER = reinforce.train(num_episodes, step_size)

    print(f'fin, duration: {(time.time() - start_time) / 60:.1f} minutes')

    fig, axs = plt.subplots(1, 2, figsize=(10, 7),)
    axs[0].plot(torch.stack(plot_ER).detach().numpy(), label='expected rewards')
    axs[0].legend()

    axs[1].plot(torch.stack(plot_total_rewards).detach().numpy(), label='total rewards')
    axs[1].legend()

    plt.savefig('Rewards.png')
    plt.show()




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.path as mpltPath
from scipy.spatial import ConvexHull
import torch

class Room:
    """
    A room in a building (2D).
    Coordinates must be entered in a clockwise or anti-clockwise order.
    """

    def __init__(self, name, coordinates):

        self.name = name
        self.coordinates = torch.tensor(coordinates)
        self.capacitance = 0  # Dummy Value, is changed during optimisation process.

        self.hull = ConvexHull(coordinates)
        self.area = self.hull.area
        self._path = mpltPath.Path(self.coordinates)

        # Iterate through sorted vertices and store each pair which form a wall
        walls = []
        for i in range(len(self.coordinates) - 1):
            walls.append((tuple(coordinates[i]), tuple(coordinates[i + 1])))
        walls.append((tuple(coordinates[-1]), tuple(coordinates[0])))
        self.walls = walls  # list of walls in room

    def pltrm(self, ax=None, colour=None):

        c = self.coordinates

        if ax is None:
            ax = plt

        line = ax.plot(
            np.hstack([c[:, 0], c[0, 0]]),
            np.hstack([c[:, 1], c[0, 1]]),
            c=colour,
        )
        return line

    def points_in_room(self, P):
        """ Boolean answering if point is inside the polygon.

        P is in form:
                P = [[x1,y1],[x2,y2]...]
        """
        return self._path.contains_points(P)



from . import room
from . import building
from . import tools
from . import RCModel
from . import optimisation
from . import reinforce


from .room import Room
from .building import Building
from .tools import InputScaling
from .tools import pltsolution_1rm
from .RCModel import RCModel
from .optimisation import OptimiseRC
from .reinforce import PolicyNetwork
from .reinforce import Reinforce
from .reinforce import LSIEnv




import pytest
import torch
import tempfile
import numpy as np

from rcmodel.building import Building
from rcmodel.RCModel import RCModel
from rcmodel.tools import InputScaling
from rcmodel.optimisation import OptimiseRC


# Function for constant outside temperature. try/except allows for broadcasting of value.
def dummy_tout(t):
    try:
        return -5 * torch.ones(len(t))

    except TypeError:
        return torch.tensor(-5)


@pytest.mark.parametrize(
    'building',
    [
        pytest.lazy_fixture('building_n2'),
        pytest.lazy_fixture('building_n9'),
    ], )
def test_run_model(building):

    # Initialise scaling methods:
    rm_CA = [200, 800]  # [min, max] Capacitance/area
    ex_C = [1.5 * 10 ** 4, 10 ** 6]  # Capacitance
    R = [0.2, 1.2]  # Resistance ((K.m^2)/W)
    Q_limit = [0, 0]  # gain limit (W/m^2) no additional energy

    scaling = InputScaling(rm_CA, ex_C, R, Q_limit)

    # Initialise RCModel with the building and InputScaling
    transform = torch.sigmoid
    model = RCModel(building, scaling, dummy_tout, transform)
    model.cooling = torch.nn.Parameter(model.cooling * 0)  # no heating

    # Set parameters for forward run:
    t_eval = torch.arange(0, 200000, 30, dtype=torch.float32)

    output = model(t_eval)

    start_temp = output[0][0, 0].item()  # initial value condition

    assert (not torch.any(output < -5).item()) and (not torch.any(output > start_temp).item()),\
        "Model has gained or lost energy from the system"


def test_save_load(building_n9):
    """
    tests if parameters are unchanged when saved and loaded with different scaling limits.
    """

    # Initialise scaling methods:
    rm_CA = [200, 800]  # [min, max] Capacitance/area
    ex_C = [1e3, 1e6]  # Capacitance
    R = [0.1, 1.3]  # Resistance ((K.m^2)/W)
    Q_limit = [-10000, 10000]  # gain limit (W/m^2)

    scaling = InputScaling(rm_CA, ex_C, R, Q_limit)
    transform = torch.sigmoid

    model = RCModel(building_n9, scaling, dummy_tout, transform)

    rm_cap = 500
    ex_cap = [1e4, 2e4]
    ex_r = [0.2, 0.5, 1.2]
    wl_r = 0.8
    gain = 2400

    original_params = torch.tensor([rm_cap, ex_cap[0], ex_cap[1], ex_r[0], ex_r[1], ex_r[2], wl_r, gain])
    original_cooling = 3000 * torch.rand(len(model.building.rooms))

    model.params = torch.nn.Parameter(torch.logit(model.scaling.model_param_scaling(original_params)))
    model.cooling = torch.nn.Parameter(torch.logit(model.scaling.model_cooling_scaling(original_cooling)))

    with tempfile.TemporaryDirectory() as tmpdirname:
        num = 0
        # save model values
        model.save(model_id=num, dir_path=tmpdirname)

        del model

        # Initialise new model with different scaling:
        rm_CA = [50, 901]
        ex_C = [1e2, 1e5]
        R = [0.2, 2]
        Q_limit = [-5200, 8100]

        scaling = InputScaling(rm_CA, ex_C, R, Q_limit)
        transform = torch.sigmoid

        model2 = RCModel(building_n9, scaling, dummy_tout, transform)

        path = f"{tmpdirname}/rcmodel{num}.pt"
        model2.load(path=path)

        loaded_params = model2.scaling.physical_param_scaling(transform(model2.params))
        loaded_cooling = model2.scaling.physical_cooling_scaling(transform(model2.cooling))

        diff_params = abs(loaded_params-original_params)
        diff_cooling = abs(loaded_cooling-original_cooling)

    assert (diff_params < 1e-3).all() and (diff_cooling < 1e-3).all(), 'model parameters are changing during a save and load'


if __name__ == '__main__':
    pytest.main()

    print("__main__ reached")
